*************************************************************************************
*														*
*	Swarm simulation for the EASy68k simulator			2007/04/18		*
*														*
*	A swarms of bees chase individual wasps round the screen. Inspired by		*
*	Xswarm.												*
*														*
*	Use [S] to toggle the status display on/off and the following key pairs ..	*
*														*
*		[B] plus [currt] to increase the bee maximum velocity				*
*		[B] plus [curlt] to decrease the bee maximum velocity				*
*		[B] plus [curup] to increase the bee maximum acceleration			*
*		[B] plus [curdn] to decrease the bee maximum acceleration			*
*														*
*		[W] plus [currt] to increase the wasp maximum velocity			*
*		[W] plus [curlt] to decrease the wasp maximum velocity			*
*		[W] plus [curup] to increase the wasp maximum acceleration			*
*		[W] plus [curdn] to decrease the wasp maximum acceleration			*
*														*
*	For bees that don't stick so tightly to the wasps increase the bee maximum	*
*	velocity , for wasps that get away more increase the wasp maximum velocity	*
*	and wasp_acceleration.										*
*														*
*	The mouse pointer has been added to the bee target list so you can attract	*
*	bees away from the wasps with the mouse.							*
*														*
*	The F2, F3 and F4	keys can be used to select a screen size of 640 x 480,	*
*	800 x 600 and 1024 x 768 respectively.							*
*														*
*	This version for Sim68K 3.6 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

screen_border	EQU	20			* screen border

number_of_wasps	EQU	4			* number of wasps
number_of_bees	EQU	20			* number of bees

wasp_colour		EQU	$000000FF		* the wasp colour
bee_colour		EQU	$0000FF00		* the bee colour

acceleration_max	EQU	10			* maximum wasp or bee acceleration
acceleration_min	EQU	2			* minimum wasp or bee acceleration
acceleration_stp	EQU	1			* wasp or bee acceleration step size

velocity_max	EQU	1000			* maximum wasp or bee velocity^2
velocity_min	EQU	100			* minimum wasp or bee velocity^2
velocity_stp	EQU	50			* wasp or bee velocity^2 step size


*************************************************************************************
*
* code start

	ORG	$1000

start
	MOVEQ		#17,d1			* enable double buffering
	MOVEQ		#92,d0			* set draw mode
	TRAP		#15

	MOVEQ		#10,d1			* OR mode drawing
	MOVEQ		#92,d0			* set draw mode
	TRAP		#15

	BSR		init				* go initialise everything

	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

main_loop
	MOVEQ		#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push the time on the stack

	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	BSR		screen_size			* test and handle widow size change keypresses
	BSR		do_keys			* test and handle bee/wasp parameter keypresses

	BSR		move_wasps			* move the wasps
	BSR		move_bees			* move the bees

	BSR		draw_status			* draw the status

	BSR		draw_wasps			* draw the wasps
	BSR		draw_bees			* draw the bees

* now wait for a while. if you have a slow machine this shouldn't slow it down much but
* if you have a very fast machine them this will stop the effect running too fast

	MOVE.l	(sp)+,d7			* restore the time in 1/100ths from stack
wait_100ms
	MOVEQ		#8,d0				* get the current time in 1/100ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract the previous time from the current
	CMP.b		#$09,d1			* compare difference with 9/100ths of a second
	BMI.s		wait_100ms			* loop if time is not up yet

* if you want to see how fast it can go comment out the BMI.s above.

	BRA		main_loop			* go round forever


*************************************************************************************
*
* check the [F2], [F3] and [F4] keys. set the screen size to 640 x 480, 800 x 600 or
* 1024 x 768 if the corresponding key has been pressed

screen_size
	MOVE.l	#$71007273,d1		* [F2], [], [F3] and [F4] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	MOVE.l	d1,d2				* copy result
	BEQ.s		notscreen			* skip screen size if no F key

	MOVE.l	#$028001E0,d1		* set 640 x 480
	TST.l		d2				* test result
	BMI		setscreen			* if F2 go set window size

	MOVE.l	#$03200258,d1		* set 800 x 600
	TST.w		d2				* test result
	BMI		setscreen			* if F3 go set window size

							* else was F4 so ..
	MOVE.l	#$04000300,d1		* set 1024 x 768
setscreen
	CMP.l		screen_width(a3),d1	* compare with current screen size
	BEQ.s		notscreen			* if already set skip setting it now

	BRA		set_window_size		* go set the window size, generate new wasps
							* and bees and return
notscreen
	RTS


*************************************************************************************
*
* test and handle keypresses. [S] toggles the status display on/off
*
* [B] plus [currt] to increase the bee maximum velocity
* [B] plus [curlt] to decrease the bee maximum velocity
* [B] plus [curup] to increase the bee maximum acceleration
* [B] plus [curdn] to decrease the bee maximum acceleration
*
* [W] plus [currt] to increase the wasp maximum velocity
* [W] plus [curlt] to decrease the wasp maximum velocity
* [W] plus [curup] to increase the wasp maximum acceleration
* [W] plus [curdn] to decrease the wasp maximum acceleration

do_keys
	MOVEQ		#$53,d1			* [][][][S] key
	MOVEQ		#0,d2				* clear longword
	MOVE.w	key_status(a3),d2		* get the last key status
	BSR		test_keys			* go test key
	BEQ		no_toggle_stat		* if no key press skip status toggle

	EOR.w		#-1,status(a3)		* toggle the status flag
no_toggle_stat
	MOVE.w	d2,key_status(a3)		* save the last key syatus

	MOVEQ		#'W',d1			* [][][][W] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	MOVE.l	d1,d2				* test the result and clear last key
	BEQ		not_wasp			* if not wasp go try bee

	MOVEQ		#0,d2				* clear longword
	MOVE.l	key_wupdnltrt(a3),d2	* get the last key status

	MOVE.w	wasp_velocity_max(a3),d5
							* get wasp acceleration
	MOVE.w	wasp_acceleration(a3),d6
							* get wasp velocity

	BSR		test_updnltrt		* test the [crup][crdn][crlt][crrt] keys
	MOVE.w	d5,wasp_velocity_max(a3)
							* save wasp acceleration
	MOVE.w	d6,wasp_acceleration(a3)
							* save wasp velocity
not_wasp
	MOVE.l	d2,key_wupdnltrt(a3)	* save the last key status

	MOVEQ		#'B',d1			* [][][][B] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	TST.b		d1				* test the result
	BEQ		not_bee			* if not bee just exit

	MOVEQ		#0,d2				* clear longword
	MOVE.l	key_bupdnltrt(a3),d2	* get the last key status

	MOVE.w	bee_velocity_max(a3),d5	* get bee acceleration
	MOVE.w	bee_acceleration(a3),d6	* get bee velocity

	BSR		test_updnltrt		* test the [crup][crdn][crlt][crrt] keys

	MOVE.w	d5,bee_velocity_max(a3)	* save bee acceleration
	MOVE.w	d6,bee_acceleration(a3)	* save bee velocity
not_bee
	MOVE.l	d2,key_bupdnltrt(a3)	* save the last key syatus
	RTS


*************************************************************************************
*
* test the [crup][crdn][crlt][crrt] keys. increase or decrease the acceleration or
* velocity, check they are in range and return the in range values. enter with
* velocity in d5.w and acceleration in d6.w, returns modified velocity in d5.w and
* acceleration in d6.w

test_updnltrt
	MOVE.l	#$26282527,d1		* [crup][crdn][crlt][crrt] keys
	BSR		test_keys			* go test key

	MOVEQ		#0,d3				* clear delta velocity
	MOVEQ		#0,d4				* clear delta acceleration

	TST.w		d1				* test [crlt] key pressed
	BPL		no_crlt			* if not [crlt] skip delta set

	MOVE.w	#-velocity_stp,d3		* set -ve delta bee velocity
no_crlt
	TST.b		d1				* test [crrt] key pressed
	BPL		no_crrt			* if not [crrt] skip delta set

	MOVE.w	#velocity_stp,d3		* set +ve delta bee velocity
no_crrt
	SWAP		d1				* [crup][crdn] to low word
	TST.w		d1				* test [crup] key pressed
	BPL		no_crup			* if not [crup] skip delta set

	MOVE.w	#acceleration_stp,d4	* set +ve delta acceleration
no_crup
	TST.b		d1				* test [crdn] key pressed
	BPL		no_crdn			* if not [crdn] skip delta set

	MOVE.w	#-acceleration_stp,d4	* set -ve delta acceleration
no_crdn
	ADD.w		d5,d3				* make new acceleration
	ADD.w		d6,d4				* make new velocity

	CMP.w		#velocity_max,d3		* compare new velocity with max
	BGT		velocity_bad		* if > max skip velocity set

	CMP.w		#velocity_min,d3		* compare new velocity with min
	BLT		velocity_bad		* if < min skip velocity set

	MOVE.w	d3,d5				* copy new velocity
velocity_bad
	CMP.w		#acceleration_max,d4	* compare new acceleration with max
	BGT		acceleration_bad		* if > max skip acceleration set

	CMP.w		#acceleration_min,d4	* compare new acceleration with min
	BLT		acceleration_bad		* if < min skip acceleration set

	MOVE.w	d4,d6				* copy new acceleration
acceleration_bad
	RTS


*************************************************************************************
*
* test key(s) d1.l, last state in d2.l. returns new pressed in d1.l and new state in
* d2.l

test_keys
	MOVE.l	d0,-(sp)			* save register
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	MOVE.l	d1,d0				* copy the result, each byte is $00 if key up,
							* $FF if key down
	EOR.l		d2,d1				* compare with last key(s) state, each byte
							* is now $FF if a key has changed or $FF if
							* a key has not changed
	MOVE.l	d0,d2				* save this result as last
	MOVE.l	(sp)+,d0			* restore register
	AND.l		d2,d1				* make each byte $FF if key just pressed or
							* $00 if key not just pressed, set Zb status
	RTS


*************************************************************************************
*
* draw the current settings

draw_status
	TST.w		status(a3)			* test the status flag
	BEQ		exit_status			* if not on just exit

	MOVE.w	#$0201,d1			* set 2,1
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	LEA		string(pc),a1		* point to string
	MOVEQ		#14,d0			* display the NULL terminated string at (a1)
	TRAP		#15

	MOVE.w	#$1601,d1			* set 22,1
	MOVE.w	bee_velocity_max(a3),d2	* maximum bee (x^2 + y^2) velocity
	BSR		display_number		* display d1.l in base 10

	MOVE.w	#$1602,d1			* set 22,2
	MOVE.w	bee_acceleration(a3),d2	* get the bee acceleration
	BSR		display_number		* display d1.l in base 10

	MOVE.w	#$1603,d1			* set 22,3
	MOVE.w	wasp_velocity_max(a3),d2
							* maximum wasp (x^2 + y^2) velocity
	BSR		display_number		* display d1.l in base 10

	MOVE.w	#$1604,d1			* set 22,4
	MOVE.w	wasp_acceleration(a3),d2
							* get the wasp acceleration
	BSR		display_number		* display d1.l in base 10

exit_status
	RTS

string
	dc.b	'Bee velocity      :       [B] [curlt]- [currt]+',$0D,$0A
	dc.b	'  Bee acceleration  :       [B] [curdn]- [curup]+',$0D,$0A
	dc.b	'  Wasp velocity     :       [W] [curlt]- [currt]+',$0D,$0A
	dc.b	'  Wasp acceleration :       [W] [curdn]- [curup]+',0
	ds.w	0


*************************************************************************************
*
* output number, cursor position in d1.w, number in d2.w

display_number
	CMP.w		#1000,d2			* compare with 1000
	BCC		print_number		* if >= 100 go output the number

	ADD.w		#$0100,d1			* else increment the x position

	CMP.w		#100,d2			* compare with 100
	BCC		print_number		* if >= 100 go output the number

	ADD.w		#$0100,d1			* else increment the x position

	CMP.w		#10,d2			* compare with 10
	BCC		print_number		* if >= 10 go output the number

	ADD.w		#$0100,d1			* else increment the x position
print_number
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	MOVEQ		#0,d1				* clear longword
	MOVE.w	d2,d1				* copy the number to display
	MOVEQ		#$0A,d2			* set base 10
	MOVEQ		#15,d0			* display d1.l in base d2.b
	TRAP		#15

	RTS


*************************************************************************************
*
* initialise everything

init
	LEA		wasp_objects(pc),a6	* wasp objects base address
	LEA		bee_objects(pc),a5	* bee objects base address
	LEA		variables(pc),a3		* variables base address

	MOVE.l	#$31415927,PRNlword(a3)	* set PRNG seed, this can be any value but 0

	MOVEQ		#0,d0				* clear longword
	MOVE.w	d0,status(a3)		* clear the show status flag
	MOVE.w	d0,key_status(a3)		* clear the status key last state
	MOVE.l	d0,key_wupdnltrt(a3)	* clear the wasp key last state
	MOVE.l	d0,key_bupdnltrt(a3)	* clear the bee key last state

	MOVE.w	#3,wasp_acceleration(a3)
							* how fast the wasps can accelerate
	MOVE.w	#5,bee_acceleration(a3)	* how fast the bees can accelerate

	MOVE.w	#400,wasp_velocity_max(a3)
							* maximum wasp (x^2 + y^2) velocity
	MOVE.w	#600,bee_velocity_max(a3)
							* maximum bee (x^2 + y^2) velocity

	MOVE.l	#$028001E0,d1		* set default window size to 640 x 480
set_window_size
	MOVE.l	d1,screen_width(a3)	* save the window size
	MOVEQ 	#33,d0			* get/set window size
	TRAP		#15


*************************************************************************************
*
* generate the wasps

make_wasps
	MOVE.w	#(number_of_wasps-1)*wasp_size,d7
							* set loop count/index
make_wasps_loop
	MOVE.w	screen_width(a3),d2	* get the screen width
	SUB.w		#(2*screen_border),d2	* - 2 * border width
	BSR		rand_d2			* generate a random number in the range 0
							* to d2.w - 1, returned in d0.w
	ADD.w		#screen_border,d0		* at least a border width in
	MOVE.w	d0,x_pos(a6,d7.w)		* set the wasp x co-ordinate

	MOVE.w	screen_height(a3),d2	* get the screen height
	SUB.w		#(2*screen_border),d2	* - 2 * border width
	BSR		rand_d2			* generate a random number in the range 0
							* to d2.w - 1, returned in d0.w
	ADD.w		#screen_border,d0		* at least a border width in
	MOVE.w	d0,y_pos(a6,d7.w)		* set the wasp y co-ordinate

	MOVE.w	wasp_acceleration(a3),d2
							* get the wasp acceleration
	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,x_vel(a6,d7.w)		* set the wasp x velocity

	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,y_vel(a6,d7.w)		* set the wasp y velocity

	SUB.w		#wasp_size,d7		* subtract one wasp structure size
	BPL		make_wasps_loop		* loop if more wasps to do



*************************************************************************************
*
* generate the bees

make_bees
	MOVE.w	#(number_of_bees-1)*bee_size,d7
							* set loop count/index
make_bees_loop
	MOVE.w	screen_width(a3),d2	* get the screen width
	SUB.w		#(2*screen_border),d2	* - 2 * border width
	BSR		rand_d2			* generate a random number in the range 0
							* to d2.w - 1, returned in d0.w
	ADD.w		#screen_border,d0		* at least a border width in
	MOVE.w	d0,x_pos(a5,d7.w)		* set the bee x co-ordinate

	MOVE.w	screen_height(a3),d2	* get the screen height
	SUB.w		#(2*screen_border),d2	* - 2 * border width
	BSR		rand_d2			* generate a random number in the range 0
							* to d2.w - 1, returned in d0.w
	ADD.w		#screen_border,d0		* at least a border width in
	MOVE.w	d0,y_pos(a5,d7.w)		* set the bee y co-ordinate

	MOVE.w	bee_acceleration(a3),d2	* get the bee acceleration
	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,x_vel(a5,d7.w)		* set the bee x velocity

	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,y_vel(a5,d7.w)		* set the bee y velocity

	SUB.w		#bee_size,d7		* subtract one bee structure size
	BPL		make_bees_loop		* loop if more bees to do

	RTS


*************************************************************************************
*
* move the wasps. the wasps wander around randomly constrained by three parameters.
* wasp_acceleration is the maximum ammount by which the x or y velocity can change
* in one go (total change can be up to SQR(2*wasp_acceleration^2), wasp_velocity_max
* is the maximum linear velocity, (x^2 + y^2 velocity), that the wasp can reach and
* lastly the screen bounds minus a border which the wasps will bounce off

move_wasps
	MOVE.w	#(number_of_wasps-1)*wasp_size,d7
							* set loop count/index
move_wasps_loop
	MOVE.w	x_pos(a6,d7.w),x_old(a6,d7.w)
							* copy current x to previous x co-ordinate
	MOVE.w	y_pos(a6,d7.w),y_old(a6,d7.w)
							* copy current y to previous y co-ordinate

							* generate the new delta velocities
	MOVE.w	wasp_acceleration(a3),d2
							* get the wasp acceleration

	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,d4				* copy the delta x velocity
	BSR		rand_pm_d2			* generate a random number in the range -(d2.w
							* - 1) to (d2.w - 1), returned in d0.w
	MOVE.w	d0,d5				* copy the delta y velocity

	MOVEQ		#0,d3				* clear longword
	MOVE.w	wasp_velocity_max(a3),d3
							* get the wasp (x^2 + y^2) velocity maximum

	LEA		(a6,d7.w),a1		* get the wasp pointer
	BSR		velocity_change		* apply the delta x,y velocity to the insect

	SUB.w		#wasp_size,d7		* subtract one wasp structure size
	BPL		move_wasps_loop		* loop if more wasps to do

	RTS


*************************************************************************************
*
* move the bees. the bees move towards the nearest wasp. this is done by trying to
* reduce SQR(delta ^2 + delta y^2) to a minimum which, as the bees are in constant
* motion, means that they effectively orbit the closest wasp.
*
* like the wasps they are constrained by three parameters. bee_acceleration is the
* maximum ammount by which the x or y velocity can change in one go (total change
* can be up to SQR(2*bee_acceleration^2), bee_velocity_max is the maximum linear
* velocity, (x^2 + y^2 velocity), that the bee can reach and lastly the screen
* bounds minus a border which the bees will bounce off

move_bees
	MOVE.w	#(number_of_bees-1)*bee_size,d7
							* set loop count/index
move_bees_loop
	MOVE.w	x_pos(a5,d7.w),x_old(a5,d7.w)
							* copy current x to previous x co-ordinate
	MOVE.w	y_pos(a5,d7.w),y_old(a5,d7.w)
							* copy current y to previous y co-ordinate

* find the distance^2 and delta x and y to the nearest wasp

	MOVE.w	y_pos(a5,d7.w),-(sp)	* stack this bee's y position
	MOVE.w	x_pos(a5,d7.w),-(sp)	* stack this bee's x position

	BSR		mouse_delta			* get delta x,y to the mouse pointer 

*##	MOVEQ		#-1,d0			* set the smallest distance^2
	MOVE.w	#(number_of_wasps-1)*wasp_size,d6
							* set loop count/index
check_wasps_loop
	MOVEM.w	(sp),d1-d2			* get this bee's x,y position

	SUB.w		x_pos(a6,d6.w),d1		* get this wasp's delta x position
	SUB.w		y_pos(a6,d6.w),d2		* get this wasp's delta y position
	MOVEM.w	d1-d2,-(sp)			* save delta x,y

	MULS.w	d1,d1				* new delta x^2
	MULS.w	d2,d2				* new delta y^2
	ADD.l		d1,d2				* new distance^2 = (delta x^2) + (delta y^2)
	CMP.l		d2,d0				* compare new distance^2 with last distance^2
	BCS		no_new_dxy2			* if new distance^2 bigger skip update

							* this wasp is closer so use the distance^2
							* and delta x, delta y to this wasp
	MOVE.l	d2,d0				* make distance^2 = new distance^2
	MOVEM.w	(sp),d4-d5			* make delta x,y = new delta x,y
no_new_dxy2
	ADDQ.l	#4,sp				* dump delta x,y from stack

	SUB.w		#wasp_size,d6		* subtract one wasp structure size
	BPL		check_wasps_loop		* loop if more wasps to check

	ADDQ.l	#4,sp				* dump this bee's x,y position from stack

	NEG.w		d4				* change the delta x and y signs so ..
	NEG.w		d5				* .. that the bee moves towards the wasp

	BSR		root_d0			* calculate SQR(distance^2)
	ADDQ.w	#1,d0				* ensure distance <> 0. this serves to remove
							* the chance of a divide by zero error and to
							* ensure the bee continually orbits the wasp

* now the distance, delta x and y to the nearest wasp are known change the x,y velocity
* towards it and move the bee

	MOVE.w	bee_acceleration(a3),d1	* set bee acceleration
	MULS.w	d1,d4				* bee acceleration * delta x
	DIVS.w	d0,d4				* bee acceleration * delta x / distance
	MULS.w	d1,d5				* bee acceleration * delta y
	DIVS.w	d0,d5				* bee acceleration * delta y / distance

	MOVEQ		#0,d3				* clear longword
	MOVE.w	bee_velocity_max(a3),d3	* set the bee (x^2 + y^2) velocity maximum

	LEA		(a5,d7.w),a1		* get the bee pointer
	BSR		velocity_change		* apply the delta x,y velocity to the insect

	SUB.w		#bee_size,d7		* subtract one bee structure size
	BPL		move_bees_loop		* loop if more bees to do

	RTS


*************************************************************************************
*
* get delta x,y to the mouse pointer

mouse_delta
*   Trap #15 Task 33 - Get/Set Output Window Size
*       D1.L High 16 bits = Width in pixels, min = 640
*            Low 16 bits  = Height in pixels, min = 480
*       D1.L = 0, get current window size as
*            High 16 bits = Width
*            Low 16 bits  = Height
	MOVEQ 	#0,d1				* get window size
	MOVEQ 	#33,d0			* get/set window size
	TRAP		#15

	MOVE.w	d1,d3				* copy screen height
	SWAP		d1				* move screen width to low word
	MOVE.w	d1,d2				* copy screen width

	MOVEQ 	#0,d1				* current mouse state
	MOVEQ 	#61,d0			* read mouse
	TRAP		#15

	MOVE.w	d1,d0				* copy mouse x
	BMI.s		off_screen			* exit if outside the output window

	CMP.w		d2,d0				* compare with screen width
	BCC.s		off_screen			* exit if outside the output window

	SWAP		d1				* move mouse y to low word
	TST.w		d1				* test mouse y
	BMI.s		off_screen			* exit if outside the output window

	CMP.w		d3,d1				* compare with screen height
	BCC.s		off_screen			* exit if outside the output window

	MOVEM.w	4(sp),d4-d5			* make delta x,y = new delta x,y
	SUB.w		d0,d4				* get this wasp's delta x position
	SUB.w		d1,d5				* get this wasp's delta y position

	MOVE.w	d4,d0				* copy delta x
	MOVE.w	d5,d1				* copy delta y

	MULS.w	d0,d0				* new delta x^2
	MULS.w	d1,d1				* new delta y^2
	ADD.l		d1,d0				* new distance^2 = (delta x^2) + (delta y^2)

	RTS


off_screen
	MOVEQ		#-1,d0			* set the smallest distance^2

	RTS


*************************************************************************************
*
* draw the wasps

draw_wasps
	MOVE.l	#wasp_colour,d1		* set the pen to the wasp colour
	MOVE.w	#(number_of_wasps-1)*wasp_size,d7
							* set loop count/index
	LEA		(a6),a1			* set the insect pointer to the wasps
	MOVE.w	#wasp_size,d6		* set wasp structure size
	BRA.s		draw_insects		* go draw the wasps and return


*************************************************************************************
*
* draw the bees

draw_bees
	MOVE.l	#bee_colour,d1		* set the pen to the bee colour
	MOVE.w	#(number_of_bees-1)*bee_size,d7
							* set loop count/index
	LEA		(a5),a1			* set the insect pointer to the bees
	MOVE.w	#bee_size,d6		* set bee structure size


*************************************************************************************
*
* draw either wasps or bees

draw_insects
	MOVEQ		#80,d0			* set the pen colour
	TRAP		#15
draw_insect_loop
	MOVE.w	x_old(a1,d7.w),d1		* get x old co-ordinate
	MOVE.w	y_old(a1,d7.w),d2		* get y old co-ordinate
	MOVE.w	x_pos(a1,d7.w),d3		* get x position co-ordinate
	MOVE.w	y_pos(a1,d7.w),d4		* get y position co-ordinate
	MOVEQ		#84,d0			* draw a line from x1,y1 to x2,y2
	TRAP		#15

	SUB.w		d6,d7				* subtract one insect structure size
	BPL		draw_insect_loop		* loop if more insects to do

	RTS


*************************************************************************************
*
* check position and return limited version. put the position parameter to check in
* d1.w, the velocity in d0.w and put the upper limit in d4.w. the limited position is
* returned in d1.w and the possibly inverted velocity component is returned in d0.w

check_position
	MOVE.l	d5,-(sp)			* save register
	MOVE.w	d0,d5				* copy the x velocity
	ADD.w		d1,d5				* add the x position
	CMP.w		#screen_border,d5		* compare lower border with new position
	BCS		check_nok			* if < lower border go do negate and add

	CMP.w		d5,d4				* compare new position with upper border
	BCC		check_ok			* if <= upper border skip velocity negate

check_nok
	NEG.w		d0				* negate the velocity
check_ok
	ADD.w		d0,d1				* add the velocity to the position
	MOVE.l	(sp)+,d5			* restore register
	RTS


*************************************************************************************
*
* apply the delta x and y velocity to the insect. the deltas are in d4.w and d5.w,
* a1 points to the bee or wasp to be changed. the maximum velocity is in d3.l.

* calculate the new velocity^2 and compare with the maximum velocity^2. If velocity^2 >
* maximum velocity^2 then scale the x and y velocities by SQR(maximun velocity^2 /
* velocity^2)

velocity_change
	MOVE.w	x_vel(a1),d0		* get the insect x velocity
	ADD.w		d4,d0				* add the delta x velocity
	MOVE.w	d0,d4				* copy the new insect x velocity
	MULS.w	d0,d0				* make new x^2

	MOVE.w	y_vel(a1),d1		* get the insect y velocity
	ADD.w		d5,d1				* add the delta y velocity
	MOVE.w	d1,d5				* copy the new insect y velocity
	MULS.w	d1,d1				* make new y^2

	ADD.l		d1,d0				* x^2 + y^2
	CMP.l		d0,d3				* compare the new velocity^2 with the maximum
	BCC		velocity_ok			* if new <= maximum then skip the scaling

* do x,y * scale where scale = SQR(max velocity^2 / new velocity^2)

	SWAP		d3				* maximum velocity^2 * $10000
	DIVU.w	d0,d3				* divide by the new velocity^2
	MOVEQ		#0,d0				* clear longword
	MOVE.w	d3,d0				* copy max velocity^2 / new velocity^2
	BSR		root_d0			* calculate SQR(max velocity^2 / new velocity^2)
	LSR.l		#1,d0				* / 2 as we're going to do signed multiplication
	SWAP		d0				* get the fractional result

	MOVE.w	d4,d3				* copy the x velocity
	BSR.s		scale_velocity		* scale it
	MOVE.w	d3,d4				* restore the x velocity
	MOVE.w	d5,d3				* copy the y velocity
	BSR.s		scale_velocity		* scale it
	MOVE.w	d3,d5				* restore the y velocity

velocity_ok
	MOVE.w	d4,d0				* copy the new x velocity
	MOVE.w	x_pos(a1),d1		* get the insect x position
	MOVE.w	screen_width(a3),d4	* get the screen width
	SUB.w		#screen_border,d4		* calculate the x max position
	BSR		check_position		* calculate and check next x position
	MOVE.w	d0,x_vel(a1)		* save the insect x velocity
	MOVE.w	d1,x_pos(a1)		* save the insect x position

	MOVE.w	d5,d0				* copy the new y velocity
	MOVE.w	y_pos(a1),d1		* get the insect y position
	MOVE.w	screen_height(a3),d4	* get the screen height
	SUB.w		#screen_border,d4		* calculate the y max position
	BSR		check_position		* calculate and check next y position
	MOVE.w	d0,y_vel(a1)		* save the insect y velocity
	MOVE.w	d1,y_pos(a1)		* save the insect y position

	RTS


*************************************************************************************
*
* scale the velocity. enter with the velocity to scale in d3. all this jumping through
* hoops to do a 16 bit unsigned * 16 bit signed multiply is needed as without it the
* cumulative errors from dropping the least significant bit become visible

scale_velocity
	MOVEQ		#0,d1				* clear longword
	MOVE.w	d3,d1				* copy the velocity
	MULS.w	d0,d3				* x velocity * scale
	ADD.l		d3,d3				* correct for earlier / 2
	TST.l		d0				* test original lowest bit
	BPL		scale_done			* skip fractional add if zero

	ADD.l		d1,d3				* add velocity
scale_done
	SWAP		d3				* high word is integer part
	RTS


*************************************************************************************
*
* perform SQR(d0.l) return the result in d0.w

root_d0
	MOVEM.l	d2-d5,-(sp)			* save registers
	MOVEQ		#0,d5				* clear root
	MOVEQ		#0,d2				* clear remainder

	MOVEQ		#$1F,d3			* $1F for DBF, 64 pairs of bits to do for
							* a 32 bit result
root_loop
	ADD.l		d0,d0				* shift highest bit of number ..
	ADDX.l	d2,d2				* .. into remainder .. never overflows
	ADD.l		d5,d5				* root = root * 2 .. never overflows
	ADD.l		d0,d0				* shift highest bit of number ..
	ADDX.l	d2,d2				* .. into remainder .. never overflows

	MOVE.l	d5,d4				* copy root
	ADD.l		d4,d4				* 2n
	ADDQ.l	#1,d4				* 2n+1

	CMP.l		d4,d2				* compare 2n+1 to remainder
	BCS.s		root_ns			* skip sub if remainder smaller

	SUB.l		d4,d2				* subtract temp from remainder
	ADDQ.l	#1,d5				* increment root
root_ns
	DBF		d3,root_loop		* loop if not all done

	SWAP		d5				* shift for correct integer result
	MOVE.l	d5,d0				* copy result
	MOVEM.l	(sp)+,d2-d5			* restore registers
	RTS


*************************************************************************************
*
* generate a random value in d0.w in range 0 to d2.w - 1

rand_d2
	BSR		random_n			* generate a random number
	MOVEQ		#0,d0				* clear longword
	MOVE.w	PRNlword(a3),d0		* get a random word
	DIVU.w	d2,d0				* divide by 2 * range
	SWAP		d0				* use remainder
	RTS


*************************************************************************************
*
* generate a random value in d0.w in range -(d2.w - 1) to +(d2.w - 1)

rand_pm_d2
	BSR		random_n			* generate a random number
	MOVE.w	PRNlword(a3),d0		* get a random word
	EXT.l		d0				* make into longword. doing this instead of
							* just getting a longword ensures the DIVS
							* never overflows
	DIVS.w	d2,d0				* divide by range
	SWAP		d0				* use remainder
	RTS


*************************************************************************************
*
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

random_n
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS


*************************************************************************************
*
* objects used

wasp_objects					* wasp objects
x_pos		EQU	*-wasp_objects		* the x co-ordinate
	ds.w	1
y_pos		EQU	*-wasp_objects		* the y co-ordinate
	ds.w	1
x_old		EQU	*-wasp_objects		* the previous x co-ordinate
	ds.w	1
y_old		EQU	*-wasp_objects		* the previous y co-ordinate
	ds.w	1
x_vel		EQU	*-wasp_objects		* the x velocity
	ds.w	1
y_vel		EQU	*-wasp_objects		* the y velocity
	ds.w	1

wasp_size	EQU	*-wasp_objects		* size of one wasp object
	ds.w	(number_of_wasps-1)*wasp_size	* space for the other wasps

bee_objects						* bee objects
	ds.w	1					* the x co-ordinate
	ds.w	1					* the y co-ordinate
	ds.w	1					* the x co-ordinate
	ds.w	1					* the y co-ordinate
	ds.w	1					* the x velocity
	ds.w	1					* the y velocity

bee_size	EQU	*-bee_objects		* size of one bee object
	ds.w	(number_of_bees-1)*bee_size	* space for the other bees


* variables used

variables						* points to variables base

	OFFSET	0				* use relative addressing

PRNlword
	ds.l	1					* PRNG seed long word
status
	ds.w	1					* show status flag
key_status
	ds.w	1					* status key last state
key_wupdnltrt
	ds.l	1					* wasp key last state
key_bupdnltrt
	ds.l	1					* bee key last state

screen_width
	ds.w	1					* screen width
screen_height
	ds.w	1					* screen height

wasp_acceleration
	ds.w	1					* how fast the wasps can accelerate
bee_acceleration
	ds.w	1					* how fast the bees can accelerate

wasp_velocity_max
	ds.w	1					* maximum wasp (x^2 + y^2) velocity
bee_velocity_max
	ds.w	1					* maximum bee (x^2 + y^2) velocity


	END	start


*************************************************************************************
